@@ -26,6 +26,8 @@
 #include <string.h>
 #include <signal.h>
 #include <unistd.h>
+#include <stdint.h>
+#include <inttypes.h>
 
 #include <event.h>
 #include <evhttp.h>
@@ -173,7 +175,7 @@ handler_stats(struct evhttp_request *request)
 
 	evbuffer_add_printf(databuf, "{");
 	evbuffer_add_printf(databuf,   "\"started\": %lu", stats.started);
-	evbuffer_add_printf(databuf, ", \"total_cmds\": %lu", stats.total_cmds);
+	evbuffer_add_printf(databuf, ", \"total_cmds\": %" PRIu64, stats.total_cmds);
 	evbuffer_add_printf(databuf, ", \"version\": \"%s\"", PACKAGE_VERSION);
 	evbuffer_add_printf(databuf, ", \"tokyocabinet_version\": \"%s\"", tcversion);
 	evbuffer_add_printf(databuf, "}\n");
@@ -191,7 +193,7 @@ handler_stats_database(struct evhttp_request *request)
 
 	evbuffer_add_printf(databuf, "{");
 	evbuffer_add_printf(databuf, "\"database\": \"%s\"", db_realpath);
-	evbuffer_add_printf(databuf, ", \"num_items\": %zu", tchdbrnum(db));
+	evbuffer_add_printf(databuf, ", \"num_items\": %" PRIu64, tchdbrnum(db));
 	evbuffer_add_printf(databuf, ", \"database_bytes\": %jd", tchdbfsiz(db));
 	evbuffer_add_printf(databuf, "}\n");
 
@@ -319,6 +321,140 @@ end:
 	++stats.total_cmds;
 }
 
+void
+handler_tree_union(struct evhttp_request *request, const char *left_key, const char *right_key, int result, int skip, int limit)
+{
+	int size;
+	int cmp_val;
+	int result_count = 0;
+
+	char* value = NULL;
+	const char* left_val = NULL;
+	const char* right_val = NULL;
+	const char* chosen_val = NULL;
+
+	TCTREE *left = NULL;
+	TCTREE *right = NULL;
+	TCTREE *chosen = NULL;
+
+	struct evbuffer *databuf = evbuffer_new();
+
+	if (limit == 0) {
+		limit = -1;
+	}
+
+	if (result == RESULT_COUNT) {
+		evbuffer_add_printf(databuf, "{\"count\": ");
+	} else {
+		evbuffer_add_printf(databuf, "{\"items\": [");
+	}
+
+	value = tchdbget(db, left_key, strlen(left_key), &size);
+	if (value) {
+		left = tctreeload(value, size, SWORDFISH_KEY_CMP, NULL);
+		free(value);
+		tctreeiterinit(left);
+		left_val = tctreeiternext2(left);
+	}
+
+	value = tchdbget(db, right_key, strlen(right_key), &size);
+	if (value) {
+		right = tctreeload(value, size, SWORDFISH_KEY_CMP, NULL);
+		free(value);
+		tctreeiterinit(right);
+		right_val = tctreeiternext2(right);
+	}
+
+	if (left_val == NULL && right_val == NULL) {
+		goto end;
+	}
+
+	// Emit once for both trees as long as they are equal
+	while (left_val || right_val)
+	{
+		if (result_count == limit)
+			break;
+
+		if (left_val && right_val) {
+			cmp_val = SWORDFISH_KEY_CMP(left_val, strlen(left_val),
+				right_val, strlen(right_val), NULL);
+		} else {
+			cmp_val = left_val ? -1 : 1;
+		}
+
+		switch ((cmp_val > 0) - (cmp_val < 0)) {
+		case 0:
+		case -1:
+			chosen = left;
+			chosen_val = left_val;
+			break;
+		case 1:
+			chosen = right;
+			chosen_val = right_val;
+			break;
+		}
+
+		/* left < right */
+		if (skip == 0) {
+			switch (result) {
+			case RESULT_KEYS:
+				if (result_count)
+					evbuffer_add_printf(databuf, ",");
+				append_json_value(databuf, chosen_val);
+				break;
+			case RESULT_VALUES:
+				if (result_count)
+					evbuffer_add_printf(databuf, ",");
+				append_json_value(databuf, tctreeget2(chosen, chosen_val));
+				break;
+			case RESULT_ALL:
+				evbuffer_add_printf(databuf,
+					(result_count == 0) ? "[" : ",[");
+				append_json_value(databuf, chosen_val);
+				evbuffer_add_printf(databuf, ",");
+				append_json_value(databuf, tctreeget2(chosen, chosen_val));
+				evbuffer_add_printf(databuf, "]");
+			}
+
+			++result_count;
+		} else {
+			/* Skip this element */
+			skip--;
+		}
+
+		switch ((cmp_val > 0) - (cmp_val < 0)) {
+		case 0:
+			/* left == right; Element intersects */
+			left_val = tctreeiternext2(left);
+			right_val = tctreeiternext2(right);
+			break;
+
+		case -1:
+			left_val = tctreeiternext2(left);
+			break;
+
+		case 1:
+			/* left > right */
+			right_val = tctreeiternext2(right);
+			break;
+		}
+	}
+end:
+	if (result == RESULT_COUNT) {
+		evbuffer_add_printf(databuf, "%d}", result_count);
+	} else {
+		evbuffer_add_printf(databuf, "]}");
+	}
+
+	REPLY_OK(request, databuf);
+
+	if (left) tctreedel(left);
+	if (right) tctreedel(right);
+
+	evbuffer_free(databuf);
+	++stats.total_cmds;
+}
+
 void
 handler_tree_difference(struct evhttp_request *request, const char *left_key, const char *right_key, int result, int skip, int limit)
 {
@@ -713,7 +849,7 @@ handler_tree_get_item(struct evhttp_request *request, const char *tree_key, cons
 		}
 
 		evbuffer_add_printf(databuf, "false");
-		REPLY_NOTFOUND(request, databuf);
+		REPLY_NOTFOUND(request, databuf, "Tree not found");
 		goto end;
 	}
 
@@ -724,7 +860,7 @@ handler_tree_get_item(struct evhttp_request *request, const char *tree_key, cons
 
 	if (!value) {
 		evbuffer_add_printf(databuf, "false");
-		REPLY_NOTFOUND(request, databuf);
+		REPLY_NOTFOUND(request, databuf, "Item not found");
 		goto end;
 	}
 
@@ -837,7 +973,7 @@ handler_database_delete(struct evhttp_request *request)
 
 	switch (request->type) {
 	case EVHTTP_REQ_POST:
-		swordfish_info("Flushing database %s\n", db_name);
+		swordfish_info("Deleting database %s\n", db_name);
 		tchdbvanish(db);
 		evbuffer_add_printf(databuf, "true");
 		REPLY_OK(request, databuf);
@@ -1116,6 +1252,7 @@ request_handler(struct evhttp_request *request, void *arg)
 	char *arg_1 = NULL;
 	char *arg_2 = NULL;
 	char *arg_3 = NULL;
+	const char *notfound_message = "Not found";
 
 	struct evkeyvalq querystr;
 	struct evbuffer *databuf = evbuffer_new();
@@ -1143,6 +1280,7 @@ request_handler(struct evhttp_request *request, void *arg)
 		database = strtok_r(NULL, "/", &saveptr);
 		if (!database) {
 			/* no database specified */
+			notfound_message = "No database specified";
 			goto notfound;
 		}
 
@@ -1168,8 +1306,6 @@ request_handler(struct evhttp_request *request, void *arg)
 				goto end;
 			}
 
-			tchdboptimize(db, 0, -1, -1, HDBTLARGE);
-
 			if (db_name != NULL)
 				free(db_name);
 
@@ -1190,6 +1326,7 @@ request_handler(struct evhttp_request *request, void *arg)
 			arg_1 = strtok_r(NULL, "/", &saveptr);
 			if (!arg_1) {
 				/* no count specified */
+				notfound_message = "No count specified";
 				goto notfound;
 			}
 
@@ -1208,6 +1345,7 @@ request_handler(struct evhttp_request *request, void *arg)
 			arg_1 = strtok_r(NULL, "/", &saveptr);
 			if (!arg_1) {
 				/* no tree specified */
+				notfound_message = "No tree specified";
 				goto notfound;
 			}
 
@@ -1223,8 +1361,10 @@ request_handler(struct evhttp_request *request, void *arg)
 				break;
 
 			case RESOURCE_ITEM:
-				if ((arg_2 = strtok_r(NULL, "/", &saveptr)) == NULL)
+				if ((arg_2 = strtok_r(NULL, "/", &saveptr)) == NULL) {
+					notfound_message = "No item specified";
 					goto notfound;
+				}
 
 				switch (request->type) {
 				case EVHTTP_REQ_POST:
@@ -1248,8 +1388,10 @@ request_handler(struct evhttp_request *request, void *arg)
 
 			case RESOURCE_INTERSECTION:
 				/* return intersection of `tree` and `arg_2` */
-				if ((arg_2 = strtok_r(NULL, "/", &saveptr)) == NULL)
+				if ((arg_2 = strtok_r(NULL, "/", &saveptr)) == NULL) {
+					notfound_message = "No second tree specified";
 					goto notfound;
+				}
 
 				arg_2 = get_typed_key(TYPE_TREE, arg_2);
 
@@ -1269,6 +1411,7 @@ request_handler(struct evhttp_request *request, void *arg)
 				default:
 					/* unknown subcommand */
 					free(arg_2);
+					notfound_message = "Unknown subcommand";
 					goto notfound;
 				}
 				free(arg_2);
@@ -1276,8 +1419,10 @@ request_handler(struct evhttp_request *request, void *arg)
 
 			case RESOURCE_DIFFERENCE:
 				/* return difference between `tree` and `arg_2` */
-				if ((arg_2 = strtok_r(NULL, "/", &saveptr)) == NULL)
+				if ((arg_2 = strtok_r(NULL, "/", &saveptr)) == NULL) {
+					notfound_message = "No second tree specified";
 					goto notfound;
+				}
 
 				arg_2 = get_typed_key(TYPE_TREE, arg_2);
 
@@ -1297,6 +1442,38 @@ request_handler(struct evhttp_request *request, void *arg)
 				default:
 					/* unknown subcommand */
 					free(arg_2);
+					notfound_message = "Unknown subcommand";
+					goto notfound;
+				}
+				free(arg_2);
+				break;
+
+			case RESOURCE_UNION:
+				/* return union between `tree` and `arg_2` */
+				if ((arg_2 = strtok_r(NULL, "/", &saveptr)) == NULL) {
+					notfound_message = "No second tree specified";
+					goto notfound;
+				}
+
+				arg_2 = get_typed_key(TYPE_TREE, arg_2);
+
+				switch (lookup_resource(strtok_r(NULL, "/", &saveptr))) {
+				case RESOURCE_NONE:
+					handler_tree_union(request, arg_1, arg_2,
+						get_values_value(&querystr),
+						get_int_header(&querystr, "skip", 0),
+						get_int_header(&querystr, "limit", 0));
+					break;
+
+				case RESOURCE_COUNT:
+					handler_tree_union(request, arg_1, arg_2,
+						RESULT_COUNT, 0, -1);
+					break;
+					
+				default:
+					/* unknown subcommand */
+					free(arg_2);
+					notfound_message = "Unknown subcommand";
 					goto notfound;
 				}
 				free(arg_2);
@@ -1316,6 +1493,7 @@ request_handler(struct evhttp_request *request, void *arg)
 
 			default:
 				/* unknown subcommand */
+				notfound_message = "Unknown tree subcommand";
 				goto notfound;
 			}
 
@@ -1323,6 +1501,7 @@ request_handler(struct evhttp_request *request, void *arg)
 
 		default:
 			/* unknown sub-database entity */
+			notfound_message = "Unknown entity";
 			goto notfound;
 		}
 
@@ -1330,13 +1509,14 @@ request_handler(struct evhttp_request *request, void *arg)
 
 	default:
 		/* unknown command */
+		notfound_message = "Unknown command";
 		goto notfound;
 	}
 
 	goto end;
 
 notfound:
-	REPLY_NOTFOUND(request, databuf);
+	REPLY_NOTFOUND(request, databuf, notfound_message);
 
 end:
 	free(uri);
@@ -1354,8 +1534,7 @@ usage(const char *progname)
 		"\t -d datadir     data directory\n"
 		"\t -i interface   interface to run server on\n"
 		"\t -P port        port number to run server on\n"
-		"\t -p pidfile     daemonise and write pid to <pidfile>\n"
-		"\t -v             enable verbose mode\n",
+		"\t -p pidfile     daemonise and write pid to <pidfile>\n",
 		progname);
 }
 
@@ -1502,7 +1681,7 @@ main(int argc, char** argv)
 			break;
 		default:
 			return EXIT_SUCCESS;
-		}	
+		}
 
 		if (setsid() == -1) {
 			perror("setsid");
